@@ -4,7 +4,7 @@ import itertools
 import string
 import typing
 
-from .. import helpers, utils, hints
+from .. import helpers, utils, hints, errors
 from ..requestiter import RequestIter
 from ..tl import types, functions, custom
 
@@ -30,13 +30,13 @@ class _ChatAction:
 
         'audio': types.SendMessageUploadAudioAction(1),
         'voice': types.SendMessageUploadAudioAction(1),  # alias
+        'song': types.SendMessageUploadAudioAction(1),  # alias
         'round': types.SendMessageUploadRoundAction(1),
         'video': types.SendMessageUploadVideoAction(1),
 
         'photo': types.SendMessageUploadPhotoAction(1),
         'document': types.SendMessageUploadDocumentAction(1),
         'file': types.SendMessageUploadDocumentAction(1),  # alias
-        'song': types.SendMessageUploadDocumentAction(1),  # alias
 
         'cancel': types.SendMessageCancelAction()
     }
@@ -337,9 +337,31 @@ class _ProfilePhotoIter(RequestIter):
             else:
                 self.request.offset += len(result.photos)
         else:
-            self.buffer = [x.action.photo for x in result.messages
-                           if isinstance(x.action, types.MessageActionChatEditPhoto)]
+            # Some broadcast channels have a photo that this request doesn't
+            # retrieve for whatever random reason the Telegram server feels.
+            #
+            # This means the `total` count may be wrong but there's not much
+            # that can be done around it (perhaps there are too many photos
+            # and this is only a partial result so it's not possible to just
+            # use the len of the result).
             self.total = getattr(result, 'count', None)
+
+            # Unconditionally fetch the full channel to obtain this photo and
+            # yield it with the rest (unless it's a duplicate).
+            seen_id = None
+            if isinstance(result, types.messages.ChannelMessages):
+                channel = await self.client(functions.channels.GetFullChannelRequest(self.request.peer))
+                photo = channel.full_chat.chat_photo
+                if isinstance(photo, types.Photo):
+                    self.buffer.append(photo)
+                    seen_id = photo.id
+
+            self.buffer.extend(
+                x.action.photo for x in result.messages
+                if isinstance(x.action, types.MessageActionChatEditPhoto)
+                and x.action.photo.id != seen_id
+            )
+
             if len(result.messages) < self.request.limit:
                 self.left = len(self.buffer)
             elif result.messages:
@@ -789,6 +811,7 @@ class ChatMethods:
             invite_users: bool = None,
             pin_messages: bool = None,
             add_admins: bool = None,
+            anonymous: bool = None,
             is_admin: bool = None,
             title: str = None) -> types.Updates:
         """
@@ -832,6 +855,18 @@ class ChatMethods:
             add_admins (`bool`, optional):
                 Whether the user will be able to add admins.
 
+            anonymous (`bool`, optional):
+                Whether the user will remain anonymous when sending messages.
+                The sender of the anonymous messages becomes the group itself.
+
+                .. note::
+
+                    Users may be able to identify the anonymous admin by its
+                    custom title, so additional care is needed when using both
+                    ``anonymous`` and custom titles. For example, if multiple
+                    anonymous admins share the same title, users won't be able
+                    to distinguish them.
+
             is_admin (`bool`, optional):
                 Whether the user will be an admin in the chat.
                 This will only work in small group chats.
@@ -871,7 +906,8 @@ class ChatMethods:
 
         perm_names = (
             'change_info', 'post_messages', 'edit_messages', 'delete_messages',
-            'ban_users', 'invite_users', 'pin_messages', 'add_admins'
+            'ban_users', 'invite_users', 'pin_messages', 'add_admins',
+            'anonymous',
         )
 
         ty = helpers._entity_type(entity)
@@ -922,6 +958,7 @@ class ChatMethods:
             send_gifs: bool = True,
             send_games: bool = True,
             send_inline: bool = True,
+            embed_link_previews: bool = True,
             send_polls: bool = True,
             change_info: bool = True,
             invite_users: bool = True,
@@ -986,6 +1023,12 @@ class ChatMethods:
             send_inline (`bool`, optional):
                 Whether the user is able to use inline bots or not.
 
+            embed_link_previews (`bool`, optional):
+                Whether the user is able to enable the link preview in the
+                messages they send. Note that the user will still be able to
+                send messages with links if this permission is removed, but
+                these links won't display a link preview.
+
             send_polls (`bool`, optional):
                 Whether the user is able to send polls or not.
 
@@ -1031,6 +1074,7 @@ class ChatMethods:
             send_gifs=not send_gifs,
             send_games=not send_games,
             send_inline=not send_inline,
+            embed_links=not embed_link_previews,
             send_polls=not send_polls,
             change_info=not change_info,
             invite_users=not invite_users,
@@ -1115,4 +1159,122 @@ class ChatMethods:
         else:
             raise ValueError('You must pass either a channel or a chat')
 
+    async def get_permissions(
+            self: 'TelegramClient',
+            entity: 'hints.EntityLike',
+            user: 'hints.EntityLike'
+    ) -> 'typing.Optional[custom.ParticipantPermissions]':
+        """
+        Fetches the permissions of a user in a specific chat or channel.
+
+        .. note::
+
+            This request has to fetch the entire chat for small group chats,
+            which can get somewhat expensive, so use of a cache is advised.
+
+        Arguments
+            entity (`entity`):
+                The channel or chat the user is participant of.
+
+            user (`entity`):
+                Target user.
+
+        Returns
+            A `ParticipantPermissions <telethon.tl.custom.participantpermissions.ParticipantPermissions>`
+            instance. Refer to its documentation to see what properties are
+            available.
+
+        Example
+            .. code-block:: python
+
+                permissions = await client.get_permissions(chat, user)
+                if permissions.is_admin:
+                    # do something
+        """
+        entity = await self.get_input_entity(entity)
+        user = await self.get_input_entity(user)
+        if helpers._entity_type(user) != helpers._EntityType.USER:
+            raise ValueError('You must pass a user entity')
+        if helpers._entity_type(entity) == helpers._EntityType.CHANNEL:
+            participant = await self(functions.channels.GetParticipantRequest(
+                entity,
+                user
+            ))
+            return custom.ParticipantPermissions(participant.participant, False)
+        elif helpers._entity_type(entity) == helpers._EntityType.CHAT:
+            chat = await self(functions.messages.GetFullChatRequest(
+                entity
+            ))
+            if isinstance(user, types.InputPeerSelf):
+                user = await self.get_me(input_peer=True)
+            for participant in chat.full_chat.participants.participants:
+                if participant.user_id == user.user_id:
+                    return custom.ParticipantPermissions(participant, True)
+            raise errors.UserNotParticipantError(None)
+
+        raise ValueError('You must pass either a channel or a chat')
+
+    async def get_stats(
+            self: 'TelegramClient',
+            entity: 'hints.EntityLike',
+    ):
+        """
+        Retrieves statistics from the given megagroup or broadcast channel.
+
+        Note that some restrictions apply before being able to fetch statistics,
+        in particular the channel must have enough members (for megagroups, this
+        requires `at least 500 members`_).
+
+        Arguments
+            entity (`entity`):
+                The channel from which to get statistics.
+
+        Raises
+            If the given entity is not a channel (broadcast or megagroup),
+            a `TypeError` is raised.
+
+            If there are not enough members (poorly named) errors such as
+            ``telethon.errors.ChatAdminRequiredError`` will appear.
+
+        Returns
+            Either :tl:`BroadcastStats` or :tl:`MegagroupStats`, depending on
+            whether the input belonged to a broadcast channel or megagroup.
+
+        Example
+            .. code-block:: python
+
+                # Some megagroup or channel username or ID to fetch
+                channel = -100123
+                stats = await client.get_stats(channel)
+                print('Stats from', stats.period.min_date, 'to', stats.period.max_date, ':')
+                print(stats.stringify())
+
+        .. _`at least 500 members`: https://telegram.org/blog/profile-videos-people-nearby-and-more
+        """
+        entity = await self.get_input_entity(entity)
+        if helpers._entity_type(entity) != helpers._EntityType.CHANNEL:
+            raise TypeError('You must pass a user entity')
+
+        # Don't bother fetching the Channel entity (costs a request), instead
+        # try to guess and if it fails we know it's the other one (best case
+        # no extra request, worst just one).
+        try:
+            req = functions.stats.GetBroadcastStatsRequest(entity)
+            return await self(req)
+        except errors.StatsMigrateError as e:
+            dc = e.dc
+        except errors.BroadcastRequiredError:
+            req = functions.stats.GetMegagroupStatsRequest(entity)
+            try:
+                return await self(req)
+            except errors.StatsMigrateError as e:
+                dc = e.dc
+
+        sender = await self._borrow_exported_sender(dc)
+        try:
+            # req will be resolved to use the right types inside by now
+            return await sender.send(req)
+        finally:
+            await self._return_exported_sender(sender)
+
     # endregion
