@@ -1,6 +1,7 @@
 import inspect
 import itertools
 import typing
+import warnings
 
 from .. import helpers, utils, errors, hints
 from ..requestiter import RequestIter
@@ -18,7 +19,7 @@ class _MessagesIter(RequestIter):
     """
     async def _init(
             self, entity, offset_id, min_id, max_id,
-            from_user, offset_date, add_offset, filter, search
+            from_user, offset_date, add_offset, filter, search, reply_to
     ):
         # Note that entity being `None` will perform a global search.
         if entity:
@@ -57,34 +58,43 @@ class _MessagesIter(RequestIter):
 
         if from_user:
             from_user = await self.client.get_input_entity(from_user)
-            ty = helpers._entity_type(from_user)
-            if ty != helpers._EntityType.USER:
-                from_user = None  # Ignore from_user unless it's a user
-
-        if from_user:
             self.from_id = await self.client.get_peer_id(from_user)
         else:
             self.from_id = None
 
-        # `messages.searchGlobal` only works with text `search` queries.
-        # If we want to perform global a search with `from_user` or `filter`,
-        # we have to perform a normal `messages.search`, *but* we can make the
-        # entity be `inputPeerEmpty`.
-        if not self.entity and (filter or from_user):
+        # `messages.searchGlobal` only works with text `search` or `filter` queries.
+        # If we want to perform global a search with `from_user` we have to perform
+        # a normal `messages.search`, *but* we can make the entity be `inputPeerEmpty`.
+        if not self.entity and from_user:
             self.entity = types.InputPeerEmpty()
 
+        if filter is None:
+            filter = types.InputMessagesFilterEmpty()
+
         if not self.entity:
             self.request = functions.messages.SearchGlobalRequest(
                 q=search or '',
-                offset_rate=offset_date,
+                filter=filter,
+                min_date=None,
+                max_date=offset_date,
+                offset_rate=None,
                 offset_peer=types.InputPeerEmpty(),
                 offset_id=offset_id,
                 limit=1
             )
+        elif reply_to is not None:
+            self.request = functions.messages.GetRepliesRequest(
+                peer=self.entity,
+                msg_id=reply_to,
+                offset_id=offset_id,
+                offset_date=offset_date,
+                add_offset=add_offset,
+                limit=1,
+                max_id=0,
+                min_id=0,
+                hash=0
+            )
         elif search is not None or filter or from_user:
-            if filter is None:
-                filter = types.InputMessagesFilterEmpty()
-
             # Telegram completely ignores `from_id` in private chats
             ty = helpers._entity_type(self.entity)
             if ty == helpers._EntityType.USER:
@@ -118,7 +128,8 @@ class _MessagesIter(RequestIter):
             #
             # Even better, using `filter` and `from_id` seems to always
             # trigger `RPC_CALL_FAIL` which is "internal issues"...
-            if filter and offset_date and not search and not offset_id:
+            if not isinstance(filter, types.InputMessagesFilterEmpty) \
+                    and offset_date and not search and not offset_id:
                 async for m in self.client.iter_messages(
                         self.entity, 1, offset_date=offset_date):
                     self.request.offset_id = m.id + 1
@@ -171,7 +182,7 @@ class _MessagesIter(RequestIter):
         messages = reversed(r.messages) if self.reverse else r.messages
         for message in messages:
             if (isinstance(message, types.MessageEmpty)
-                    or self.from_id and message.from_id != self.from_id):
+                    or self.from_id and message.sender_id != self.from_id):
                 continue
 
             if not self._message_in_range(message):
@@ -191,7 +202,7 @@ class _MessagesIter(RequestIter):
         # Get the last message that's not empty (in some rare cases
         # it can happen that the last message is :tl:`MessageEmpty`)
         if self.buffer:
-            self._update_offset(self.buffer[-1])
+            self._update_offset(self.buffer[-1], r)
         else:
             # There are some cases where all the messages we get start
             # being empty. This can happen on migrated mega-groups if
@@ -217,7 +228,7 @@ class _MessagesIter(RequestIter):
 
         return True
 
-    def _update_offset(self, last_message):
+    def _update_offset(self, last_message, response):
         """
         After making the request, update its offset with the last message.
         """
@@ -233,11 +244,16 @@ class _MessagesIter(RequestIter):
             # (only for the first request), it's safe to just clear it off.
             self.request.max_date = None
         else:
-            # getHistory and searchGlobal call it offset_date
+            # getHistory, searchGlobal and getReplies call it offset_date
             self.request.offset_date = last_message.date
 
         if isinstance(self.request, functions.messages.SearchGlobalRequest):
-            self.request.offset_peer = last_message.input_chat
+            if last_message.input_chat:
+                self.request.offset_peer = last_message.input_chat
+            else:
+                self.request.offset_peer = types.InputPeerEmpty()
+
+            self.request.offset_rate = getattr(response, 'next_rate', None)
 
 
 class _IDsIter(RequestIter):
@@ -270,7 +286,7 @@ class _IDsIter(RequestIter):
         else:
             r = await self.client(functions.messages.GetMessagesRequest(ids))
             if self._entity:
-                from_id = await self.client.get_peer_id(self._entity)
+                from_id = await self.client._get_peer(self._entity)
 
         if isinstance(r, types.messages.MessagesNotModified):
             self.buffer.extend(None for _ in ids)
@@ -289,7 +305,7 @@ class _IDsIter(RequestIter):
         # arbitrary chats. Validate these unless ``from_id is None``.
         for message in r.messages:
             if isinstance(message, types.MessageEmpty) or (
-                    from_id and message.chat_id != from_id):
+                    from_id and message.peer_id != from_id):
                 self.buffer.append(None)
             else:
                 message._finish_init(self.client, entities, self._entity)
@@ -317,7 +333,8 @@ class MessageMethods:
             from_user: 'hints.EntityLike' = None,
             wait_time: float = None,
             ids: 'typing.Union[int, typing.Sequence[int]]' = None,
-            reverse: bool = False
+            reverse: bool = False,
+            reply_to: int = None
     ) -> 'typing.Union[_MessagesIter, _IDsIter]':
         """
         Iterator over the messages for the given chat.
@@ -384,8 +401,7 @@ class MessageMethods:
                 containing photos.
 
             from_user (`entity`):
-                Only messages from this user will be returned.
-                This parameter will be ignored if it is not an user.
+                Only messages from this entity will be returned.
 
             wait_time (`int`):
                 Wait time (in seconds) between different
@@ -425,6 +441,26 @@ class MessageMethods:
 
                 You cannot use this if both `entity` and `ids` are `None`.
 
+            reply_to (`int`, optional):
+                If set to a message ID, the messages that reply to this ID
+                will be returned. This feature is also known as comments in
+                posts of broadcast channels, or viewing threads in groups.
+
+                This feature can only be used in broadcast channels and their
+                linked megagroups. Using it in a chat or private conversation
+                will result in ``telethon.errors.PeerIdInvalidError`` to occur.
+
+                When using this parameter, the ``filter`` and ``search``
+                parameters have no effect, since Telegram's API doesn't
+                support searching messages in replies.
+
+                .. note::
+
+                    This feature is used to get replies to a message in the
+                    *discussion* group. If the same broadcast channel sends
+                    a message and replies to it itself, that reply will not
+                    be included in the results.
+
         Yields
             Instances of `Message <telethon.tl.custom.message.Message>`.
 
@@ -451,6 +487,10 @@ class MessageMethods:
                 from telethon.tl.types import InputMessagesFilterPhotos
                 async for message in client.iter_messages(chat, filter=InputMessagesFilterPhotos):
                     print(message.photo)
+
+                # Getting comments from a post in a channel:
+                async for message in client.iter_messages(channel, reply_to=123):
+                    print(message.chat.title, message.text)
         """
         if ids is not None:
             if not utils.is_list_like(ids):
@@ -478,7 +518,8 @@ class MessageMethods:
             offset_date=offset_date,
             add_offset=add_offset,
             filter=filter,
-            search=search
+            search=search,
+            reply_to=reply_to
         )
 
     async def get_messages(self: 'TelegramClient', *args, **kwargs) -> 'hints.TotalList':
@@ -543,6 +584,7 @@ class MessageMethods:
             *,
             reply_to: 'typing.Union[int, types.Message]' = None,
             parse_mode: typing.Optional[str] = (),
+            formatting_entities: typing.Optional[typing.List[types.TypeMessageEntity]] = None,
             link_preview: bool = True,
             file: 'typing.Union[hints.FileLike, typing.Sequence[hints.FileLike]]' = None,
             force_document: bool = False,
@@ -590,6 +632,9 @@ class MessageMethods:
                 property for allowed values. Markdown parsing will be used by
                 default.
 
+            formatting_entities (`list`, optional):
+                A list of message formatting entities. When provided, the ``parse_mode`` is ignored.
+
             link_preview (`bool`, optional):
                 Should the link preview be shown?
 
@@ -602,7 +647,6 @@ class MessageMethods:
 
             clear_draft (`bool`, optional):
                 Whether the existing draft should be cleared or not.
-                Has no effect when sending a file.
 
             buttons (`list`, `custom.Button <telethon.tl.custom.button.Button>`, :tl:`KeyboardButton`):
                 The matrix (list of lists), row list or button to be shown
@@ -635,7 +679,7 @@ class MessageMethods:
             .. code-block:: python
 
                 # Markdown is the default
-                await client.send_message('lonami', 'Thanks for the **Telethon** library!')
+                await client.send_message('me', 'Hello **world**!')
 
                 # Default to another parse mode
                 client.parse_mode = 'html'
@@ -669,7 +713,7 @@ class MessageMethods:
                 # Matrix of inline buttons
                 await client.send_message(chat, 'Pick one from this grid', buttons=[
                     [Button.inline('Left'), Button.inline('Right')],
-                    [Button.url('Check this site!', 'https://lonamiwebs.github.io')]
+                    [Button.url('Check this site!', 'https://example.com')]
                 ])
 
                 # Reply keyboard
@@ -691,7 +735,8 @@ class MessageMethods:
             return await self.send_file(
                 entity, file, caption=message, reply_to=reply_to,
                 parse_mode=parse_mode, force_document=force_document,
-                buttons=buttons
+                buttons=buttons, clear_draft=clear_draft, silent=silent,
+                schedule=schedule, formatting_entities=formatting_entities
             )
 
         entity = await self.get_input_entity(entity)
@@ -713,7 +758,7 @@ class MessageMethods:
                     silent=silent,
                     reply_to=reply_to,
                     buttons=markup,
-                    entities=message.entities,
+                    formatting_entities=message.entities,
                     schedule=schedule
                 )
 
@@ -731,7 +776,8 @@ class MessageMethods:
             )
             message = message.message
         else:
-            message, msg_ent = await self._parse_message_text(message, parse_mode)
+            if formatting_entities is None:
+                message, formatting_entities = await self._parse_message_text(message, parse_mode)
             if not message:
                 raise ValueError(
                     'The message cannot be empty unless a file is provided'
@@ -740,7 +786,7 @@ class MessageMethods:
             request = functions.messages.SendMessageRequest(
                 peer=entity,
                 message=message,
-                entities=msg_ent,
+                entities=formatting_entities,
                 no_webpage=not link_preview,
                 reply_to_msg_id=utils.get_message_id(reply_to),
                 clear_draft=clear_draft,
@@ -753,7 +799,7 @@ class MessageMethods:
         if isinstance(result, types.UpdateShortSentMessage):
             message = types.Message(
                 id=result.id,
-                to_id=utils.get_peer(entity),
+                peer_id=await self._get_peer(entity),
                 message=message,
                 date=result.date,
                 out=result.out,
@@ -805,15 +851,7 @@ class MessageMethods:
                 this behaviour.
 
             as_album (`bool`, optional):
-                Whether several image messages should be forwarded as an
-                album (grouped) or not. The default behaviour is to treat
-                albums specially and send outgoing requests with
-                ``as_album=True`` only for the albums if message objects
-                are used. If IDs are used it will group by default.
-
-                In short, the default should do what you expect,
-                `True` will group always (even converting separate
-                images into albums), and `False` will never group.
+                This flag no longer has any effect.
 
             schedule (`hints.DateLike`, optional):
                 If set, the message(s) won't forward immediately, and
@@ -846,6 +884,9 @@ class MessageMethods:
                 # Forwarding as a copy
                 await client.send_message(chat, message)
         """
+        if as_album is not None:
+            warnings.warn('the as_album argument is deprecated and no longer has any effect')
+
         single = not utils.is_list_like(messages)
         if single:
             messages = (messages,)
@@ -858,44 +899,24 @@ class MessageMethods:
         else:
             from_peer_id = None
 
-        def _get_key(m):
+        def get_key(m):
             if isinstance(m, int):
                 if from_peer_id is not None:
-                    return from_peer_id, None
+                    return from_peer_id
 
                 raise ValueError('from_peer must be given if integer IDs are used')
             elif isinstance(m, types.Message):
-                return m.chat_id, m.grouped_id
+                return m.chat_id
             else:
                 raise TypeError('Cannot forward messages of type {}'.format(type(m)))
 
-        # We want to group outgoing chunks differently if we are "smart"
-        # about sending as album.
-        #
-        # Why? We need separate requests for ``as_album=True/False``, so
-        # if we want that behaviour, when we group messages to create the
-        # chunks, we need to consider the grouped ID too. But if we don't
-        # care about that, we don't need to consider it for creating the
-        # chunks, so we can make less requests.
-        if as_album is None:
-            get_key = _get_key
-        else:
-            def get_key(m):
-                return _get_key(m)[0]  # Ignore grouped_id
-
         sent = []
-        for chat_id, chunk in itertools.groupby(messages, key=get_key):
+        for _chat_id, chunk in itertools.groupby(messages, key=get_key):
             chunk = list(chunk)
             if isinstance(chunk[0], int):
                 chat = from_peer
-                grouped = True if as_album is None else as_album
             else:
                 chat = await chunk[0].get_input_chat()
-                if as_album is None:
-                    grouped = any(m.grouped_id is not None for m in chunk)
-                else:
-                    grouped = as_album
-
                 chunk = [m.id for m in chunk]
 
             req = functions.messages.ForwardMessagesRequest(
@@ -903,10 +924,6 @@ class MessageMethods:
                 id=chunk,
                 to_peer=entity,
                 silent=silent,
-                # Trying to send a single message as grouped will cause
-                # GROUPED_MEDIA_INVALID. If more than one message is forwarded
-                # (even without media...), this error goes away.
-                grouped=len(chunk) > 1 and grouped,
                 schedule_date=schedule
             )
             result = await self(req)
@@ -921,6 +938,7 @@ class MessageMethods:
             text: str = None,
             *,
             parse_mode: str = (),
+            formatting_entities: typing.Optional[typing.List[types.TypeMessageEntity]] = None,
             link_preview: bool = True,
             file: 'hints.FileLike' = None,
             force_document: bool = False,
@@ -960,6 +978,9 @@ class MessageMethods:
                 property for allowed values. Markdown parsing will be used by
                 default.
 
+            formatting_entities (`list`, optional):
+                A list of message formatting entities. When provided, the ``parse_mode`` is ignored.
+
             link_preview (`bool`, optional):
                 Should the link preview be shown?
 
@@ -1018,21 +1039,33 @@ class MessageMethods:
         elif isinstance(entity, types.Message):
             text = message  # Shift the parameters to the right
             message = entity
-            entity = entity.to_id
+            entity = entity.peer_id
 
-        text, msg_entities = await self._parse_message_text(text, parse_mode)
+        if formatting_entities is None:
+            text, formatting_entities = await self._parse_message_text(text, parse_mode)
         file_handle, media, image = await self._file_to_media(file,
                 force_document=force_document)
 
         if isinstance(entity, types.InputBotInlineMessageID):
-            return await self(functions.messages.EditInlineBotMessageRequest(
+            request = functions.messages.EditInlineBotMessageRequest(
                 id=entity,
                 message=text,
                 no_webpage=not link_preview,
-                entities=msg_entities,
+                entities=formatting_entities,
                 media=media,
                 reply_markup=self.build_reply_markup(buttons)
-            ))
+            )
+            # Invoke `messages.editInlineBotMessage` from the right datacenter.
+            # Otherwise, Telegram will error with `MESSAGE_ID_INVALID` and do nothing.
+            exported = self.session.dc_id != entity.dc_id
+            if exported:
+                try:
+                    sender = await self._borrow_exported_sender(entity.dc_id)
+                    return await self._call(sender, request)
+                finally:
+                    await self._return_exported_sender(sender)
+            else:
+                return await self(request)
 
         entity = await self.get_input_entity(entity)
         request = functions.messages.EditMessageRequest(
@@ -1040,13 +1073,12 @@ class MessageMethods:
             id=utils.get_message_id(message),
             message=text,
             no_webpage=not link_preview,
-            entities=msg_entities,
+            entities=formatting_entities,
             media=media,
             reply_markup=self.build_reply_markup(buttons),
             schedule_date=schedule
         )
         msg = self._get_response_message(request, await self(request), entity)
-        await self._cache_media(msg, file, file_handle, image=image)
         return msg
 
     async def delete_messages(
@@ -1108,8 +1140,14 @@ class MessageMethods:
             else int(m) for m in message_ids
         )
 
-        entity = await self.get_input_entity(entity) if entity else None
-        if helpers._entity_type(entity) == helpers._EntityType.CHANNEL:
+        if entity:
+            entity = await self.get_input_entity(entity)
+            ty = helpers._entity_type(entity)
+        else:
+            # no entity (None), set a value that's not a channel for private delete
+            ty = helpers._EntityType.USER
+
+        if ty == helpers._EntityType.CHANNEL:
             return await self([functions.channels.DeleteMessagesRequest(
                          entity, list(c)) for c in utils.chunks(message_ids)])
         else:
@@ -1136,6 +1174,10 @@ class MessageMethods:
         If neither message nor maximum ID are provided, all messages will be
         marked as read by assuming that ``max_id = 0``.
 
+        If a message or maximum ID is provided, all the messages up to and
+        including such ID will be marked as read (for all messages whose ID
+        ≤ max_id).
+
         See also `Message.mark_read() <telethon.tl.custom.message.Message.mark_read>`.
 
         Arguments
@@ -1146,8 +1188,8 @@ class MessageMethods:
                 Either a list of messages or a single message.
 
             max_id (`int`):
-                Overrides messages, until which message should the
-                acknowledge should be sent.
+                Until which message should the read acknowledge be sent for.
+                This has priority over the ``message`` parameter.
 
             clear_mentions (`bool`):
                 Whether the mention badge should be cleared (so that
@@ -1199,7 +1241,7 @@ class MessageMethods:
             notify: bool = False
     ):
         """
-        Pins or unpins a message in a chat.
+        Pins a message in a chat.
 
         The default behaviour is to *not* notify members, unlike the
         official applications.
@@ -1212,7 +1254,7 @@ class MessageMethods:
 
             message (`int` | `Message <telethon.tl.custom.message.Message>`):
                 The message or the message ID to pin. If it's
-                `None`, the message will be unpinned instead.
+                `None`, all messages will be unpinned instead.
 
             notify (`bool`, optional):
                 Whether the pin should notify people or not.
@@ -1224,13 +1266,63 @@ class MessageMethods:
                 message = await client.send_message(chat, 'Pinotifying is fun!')
                 await client.pin_message(chat, message, notify=True)
         """
+        return await self._pin(entity, message, unpin=False, notify=notify)
+
+    async def unpin_message(
+            self: 'TelegramClient',
+            entity: 'hints.EntityLike',
+            message: 'typing.Optional[hints.MessageIDLike]' = None,
+            *,
+            notify: bool = False
+    ):
+        """
+        Unpins a message in a chat.
+
+        If no message ID is specified, all pinned messages will be unpinned.
+
+        See also `Message.unpin() <telethon.tl.custom.message.Message.unpin>`.
+
+        Arguments
+            entity (`entity`):
+                The chat where the message should be pinned.
+
+            message (`int` | `Message <telethon.tl.custom.message.Message>`):
+                The message or the message ID to unpin. If it's
+                `None`, all messages will be unpinned instead.
+
+        Example
+            .. code-block:: python
+
+                # Unpin all messages from a chat
+                await client.unpin_message(chat)
+        """
+        return await self._pin(entity, message, unpin=True, notify=notify)
+
+    async def _pin(self, entity, message, *, unpin, notify=False):
         message = utils.get_message_id(message) or 0
         entity = await self.get_input_entity(entity)
-        await self(functions.messages.UpdatePinnedMessageRequest(
+        if message <= 0:  # old behaviour accepted negative IDs to unpin
+            await self(functions.messages.UnpinAllMessagesRequest(entity))
+            return
+
+        request = functions.messages.UpdatePinnedMessageRequest(
             peer=entity,
             id=message,
-            silent=not notify
-        ))
+            silent=not notify,
+            unpin=unpin,
+        )
+        result = await self(request)
+
+        # Unpinning does not produce a service message
+        if unpin:
+            return
+
+        # Pinning in User chats (just with yourself really) does not produce a service message
+        if helpers._entity_type(entity) == helpers._EntityType.USER:
+            return
+
+        # Pinning a message that doesn't exist would RPC-error earlier
+        return self._get_response_message(request, result, entity)
 
     # endregion
 
