@@ -1,4 +1,5 @@
 import asyncio
+import inspect
 import itertools
 import random
 import time
@@ -325,7 +326,7 @@ class UpdateMethods:
         while self.is_connected():
             try:
                 await asyncio.wait_for(
-                    self.disconnected, timeout=60, loop=self._loop
+                    self.disconnected, timeout=60
                 )
                 continue  # We actually just want to act upon timeout
             except asyncio.TimeoutError:
@@ -335,6 +336,9 @@ class UpdateMethods:
             except Exception:
                 continue  # Any disconnected exception should be ignored
 
+            # Check if we have any exported senders to clean-up periodically
+            await self._clean_exported_senders()
+
             # Don't bother sending pings until the low-level connection is
             # ready, otherwise a lot of pings will be batched to be sent upon
             # reconnect, when we really don't care about that.
@@ -388,12 +392,28 @@ class UpdateMethods:
                     await self._get_difference(update, channel_id, pts_date)
                 except OSError:
                     pass  # We were disconnected, that's okay
+                except errors.RPCError:
+                    # There's a high chance the request fails because we lack
+                    # the channel. Because these "happen sporadically" (#1428)
+                    # we should be okay (no flood waits) even if more occur.
+                    pass
+                except ValueError:
+                    # There is a chance that GetFullChannelRequest and GetDifferenceRequest
+                    # inside the _get_difference() function will end up with 
+                    # ValueError("Request was unsuccessful N time(s)") for whatever reasons.
+                    pass
 
         if not self._self_input_peer:
             # Some updates require our own ID, so we must make sure
             # that the event builder has offline access to it. Calling
             # `get_me()` will cache it under `self._self_input_peer`.
-            await self.get_me(input_peer=True)
+            #
+            # It will return `None` if we haven't logged in yet which is
+            # fine, we will just retry next time anyway.
+            try:
+                await self.get_me(input_peer=True)
+            except OSError:
+                pass  # might not have connection
 
         built = EventBuilderDict(self, update, others)
         for conv_set in self._conversations.values():
@@ -421,7 +441,52 @@ class UpdateMethods:
             if not builder.resolved:
                 await builder.resolve(self)
 
-            if not builder.filter(event):
+            filter = builder.filter(event)
+            if inspect.isawaitable(filter):
+                filter = await filter
+            if not filter:
+                continue
+
+            try:
+                await callback(event)
+            except errors.AlreadyInConversationError:
+                name = getattr(callback, '__name__', repr(callback))
+                self._log[__name__].debug(
+                    'Event handler "%s" already has an open conversation, '
+                    'ignoring new one', name)
+            except events.StopPropagation:
+                name = getattr(callback, '__name__', repr(callback))
+                self._log[__name__].debug(
+                    'Event handler "%s" stopped chain of propagation '
+                    'for event %s.', name, type(event).__name__
+                )
+                break
+            except Exception as e:
+                if not isinstance(e, asyncio.CancelledError) or self.is_connected():
+                    name = getattr(callback, '__name__', repr(callback))
+                    self._log[__name__].exception('Unhandled exception on %s',
+                                                  name)
+
+    async def _dispatch_event(self: 'TelegramClient', event):
+        """
+        Dispatches a single, out-of-order event. Used by `AlbumHack`.
+        """
+        # We're duplicating a most logic from `_dispatch_update`, but all in
+        # the name of speed; we don't want to make it worse for all updates
+        # just because albums may need it.
+        for builder, callback in self._event_builders:
+            if isinstance(builder, events.Raw):
+                continue
+            if not isinstance(event, builder.Event):
+                continue
+
+            if not builder.resolved:
+                await builder.resolve(self)
+
+            filter = builder.filter(event)
+            if inspect.isawaitable(filter):
+                filter = await filter
+            if not filter:
                 continue
 
             try:
@@ -456,8 +521,13 @@ class UpdateMethods:
         self._log[__name__].debug('Getting difference for entities '
                                   'for %r', update.__class__)
         if channel_id:
+            # There are reports where we somehow call get channel difference
+            # with `InputPeerEmpty`. Check our assumptions to better debug
+            # this when it happens.
+            assert isinstance(channel_id, int), 'channel_id was {}, not int in {}'.format(type(channel_id), update)
             try:
-                where = await self.get_input_entity(channel_id)
+                # Wrap the ID inside a peer to ensure we get a channel back.
+                where = await self.get_input_entity(types.PeerChannel(channel_id))
             except ValueError:
                 # There's a high chance that this fails, since
                 # we are getting the difference to fetch entities.
@@ -502,6 +572,14 @@ class UpdateMethods:
 
     async def _handle_auto_reconnect(self: 'TelegramClient'):
         # TODO Catch-up
+        # For now we make a high-level request to let Telegram
+        # know we are still interested in receiving more updates.
+        try:
+            await self.get_me()
+        except Exception as e:
+            self._log[__name__].warning('Error executing high-level request '
+                                        'after reconnect: %s: %s', type(e), e)
+
         return
         try:
             self._log[__name__].info(
@@ -552,7 +630,7 @@ class EventBuilderDict:
             return self.__dict__[builder]
         except KeyError:
             event = self.__dict__[builder] = builder.build(
-                self.update, self.others, self.client._self_input_peer.user_id)
+                self.update, self.others, self.client._self_id)
 
             if isinstance(event, EventCommon):
                 event.original_update = self.update
