@@ -27,6 +27,9 @@ def _fmt_flood(delay, request, *, early=False, td=datetime.timedelta):
 
 class UserMethods:
     async def __call__(self: 'TelegramClient', request, ordered=False):
+        return await self._call(self._sender, request, ordered=ordered)
+
+    async def _call(self: 'TelegramClient', sender, request, ordered=False):
         requests = (request if utils.is_list_like(request) else (request,))
         for r in requests:
             if not isinstance(r, TLRequest):
@@ -41,16 +44,18 @@ class UserMethods:
                     self._flood_waited_requests.pop(r.CONSTRUCTOR_ID, None)
                 elif diff <= self.flood_sleep_threshold:
                     self._log[__name__].info(*_fmt_flood(diff, r, early=True))
-                    await asyncio.sleep(diff, loop=self._loop)
+                    await asyncio.sleep(diff)
                     self._flood_waited_requests.pop(r.CONSTRUCTOR_ID, None)
                 else:
                     raise errors.FloodWaitError(request=r, capture=diff)
 
         request_index = 0
+        last_error = None
         self._last_request = time.time()
+
         for attempt in retry_range(self._request_retries):
             try:
-                future = self._sender.send(request, ordered=ordered)
+                future = sender.send(request, ordered=ordered)
                 if isinstance(future, list):
                     results = []
                     exceptions = []
@@ -76,26 +81,37 @@ class UserMethods:
                     self._entity_cache.add(result)
                     return result
             except (errors.ServerError, errors.RpcCallFailError,
-                    errors.RpcMcgetFailError) as e:
+                    errors.RpcMcgetFailError, errors.InterdcCallErrorError,
+                    errors.InterdcCallRichErrorError) as e:
+                last_error = e
                 self._log[__name__].warning(
                     'Telegram is having internal issues %s: %s',
                     e.__class__.__name__, e)
 
                 await asyncio.sleep(2)
             except (errors.FloodWaitError, errors.SlowModeWaitError, errors.FloodTestPhoneWaitError) as e:
+                last_error = e
                 if utils.is_list_like(request):
                     request = request[request_index]
 
-                self._flood_waited_requests\
-                    [request.CONSTRUCTOR_ID] = time.time() + e.seconds
+                # SLOW_MODE_WAIT is chat-specific, not request-specific
+                if not isinstance(e, errors.SlowModeWaitError):
+                    self._flood_waited_requests\
+                        [request.CONSTRUCTOR_ID] = time.time() + e.seconds
+
+                # In test servers, FLOOD_WAIT_0 has been observed, and sleeping for
+                # such a short amount will cause retries very fast leading to issues.
+                if e.seconds == 0:
+                    e.seconds = 1
 
                 if e.seconds <= self.flood_sleep_threshold:
                     self._log[__name__].info(*_fmt_flood(e.seconds, request))
-                    await asyncio.sleep(e.seconds, loop=self._loop)
+                    await asyncio.sleep(e.seconds)
                 else:
                     raise
             except (errors.PhoneMigrateError, errors.NetworkMigrateError,
                     errors.UserMigrateError) as e:
+                last_error = e
                 self._log[__name__].info('Phone migrated to %d', e.new_dc)
                 should_raise = isinstance(e, (
                     errors.PhoneMigrateError, errors.NetworkMigrateError
@@ -104,6 +120,8 @@ class UserMethods:
                     raise
                 await self._switch_dc(e.new_dc)
 
+        if self._raise_last_call_error and last_error is not None:
+            raise last_error
         raise ValueError('Request was unsuccessful {} time(s)'
                          .format(attempt))
 
@@ -148,6 +166,16 @@ class UserMethods:
         except errors.UnauthorizedError:
             return None
 
+    @property
+    def _self_id(self: 'TelegramClient') -> typing.Optional[int]:
+        """
+        Returns the ID of the logged-in user, if known.
+
+        This property is used in every update, and some like `updateLoginToken`
+        occur prior to login, so it gracefully handles when no ID is known yet.
+        """
+        return self._self_input_peer.user_id if self._self_input_peer else None
+
     async def is_bot(self: 'TelegramClient') -> bool:
         """
         Return `True` if the signed-in user is a bot, `False` otherwise.
@@ -281,7 +309,7 @@ class UserMethods:
             users = tmp
         if chats:  # TODO Handle chats slice?
             chats = (await self(
-                functions.messages.GetChatsRequest(chats))).chats
+                functions.messages.GetChatsRequest([x.chat_id for x in chats]))).chats
         if channels:
             channels = (await self(
                 functions.channels.GetChannelsRequest(channels))).chats
@@ -437,6 +465,10 @@ class UserMethods:
             .format(peer)
         )
 
+    async def _get_peer(self: 'TelegramClient', peer: 'hints.EntityLike'):
+        i, cls = utils.resolve_id(await self.get_peer_id(peer))
+        return cls(i)
+
     async def get_peer_id(
             self: 'TelegramClient',
             peer: 'hints.EntityLike',
