@@ -1,5 +1,6 @@
 import asyncio
 import collections
+import struct
 
 from . import authenticator
 from ..extensions.messagepacker import MessagePacker
@@ -39,12 +40,11 @@ class MTProtoSender:
     A new authorization key will be generated on connection if no other
     key exists yet.
     """
-    def __init__(self, auth_key, loop, *, loggers,
+    def __init__(self, auth_key, *, loggers,
                  retries=5, delay=1, auto_reconnect=True, connect_timeout=None,
                  auth_key_callback=None,
                  update_callback=None, auto_reconnect_callback=None):
         self._connection = None
-        self._loop = loop
         self._loggers = loggers
         self._log = loggers[__name__]
         self._retries = retries
@@ -54,7 +54,7 @@ class MTProtoSender:
         self._auth_key_callback = auth_key_callback
         self._update_callback = update_callback
         self._auto_reconnect_callback = auto_reconnect_callback
-        self._connect_lock = asyncio.Lock(loop=loop)
+        self._connect_lock = asyncio.Lock()
 
         # Whether the user has explicitly connected or disconnected.
         #
@@ -64,7 +64,7 @@ class MTProtoSender:
         # pending futures should be cancelled.
         self._user_connected = False
         self._reconnecting = False
-        self._disconnected = self._loop.create_future()
+        self._disconnected = asyncio.get_event_loop().create_future()
         self._disconnected.set_result(None)
 
         # We need to join the loops upon disconnection
@@ -77,8 +77,7 @@ class MTProtoSender:
 
         # Outgoing messages are put in a queue and sent in a batch.
         # Note that here we're also storing their ``_RequestState``.
-        self._send_queue = MessagePacker(self._state, self._loop,
-                                         loggers=self._loggers)
+        self._send_queue = MessagePacker(self._state, loggers=self._loggers)
 
         # Sent states are remembered until a response is received.
         self._pending_state = {}
@@ -169,7 +168,14 @@ class MTProtoSender:
             raise ConnectionError('Cannot send requests while disconnected')
 
         if not utils.is_list_like(request):
-            state = RequestState(request, self._loop)
+            try:
+                state = RequestState(request)
+            except struct.error as e:
+                # "struct.error: required argument is not an integer" is not
+                # very helpful; log the request to find out what wasn't int.
+                self._log.error('Request caused struct.error: %s: %s', e, request)
+                raise
+
             self._send_queue.append(state)
             return state.future
         else:
@@ -177,7 +183,12 @@ class MTProtoSender:
             futures = []
             state = None
             for req in request:
-                state = RequestState(req, self._loop, after=ordered and state)
+                try:
+                    state = RequestState(req, after=ordered and state)
+                except struct.error as e:
+                    self._log.error('Request caused struct.error: %s: %s', e, request)
+                    raise
+
                 states.append(state)
                 futures.append(state.future)
 
@@ -193,7 +204,7 @@ class MTProtoSender:
         Note that it may resolve in either a ``ConnectionError``
         or any other unexpected error that could not be handled.
         """
-        return asyncio.shield(self._disconnected, loop=self._loop)
+        return asyncio.shield(self._disconnected)
 
     # Private methods
 
@@ -206,6 +217,7 @@ class MTProtoSender:
         self._log.info('Connecting to %s...', self._connection)
 
         connected = False
+        
         for attempt in retry_range(self._retries):
             if not connected:
                 connected = await self._try_connect(attempt)
@@ -228,29 +240,30 @@ class MTProtoSender:
                     # reconnect cleanly after.
                     await self._connection.disconnect()
                     connected = False
-                    await asyncio.sleep(self._delay, loop=self._loop)
+                    await asyncio.sleep(self._delay)
                     continue  # next iteration we will try to reconnect
 
             break  # all steps done, break retry loop
         else:
             if not connected:
-                raise ConnectionError('Connection to Telegram failed %d time(s)', self._retries)
+                raise ConnectionError('Connection to Telegram failed {} time(s)'.format(self._retries))
 
-            e = ConnectionError('auth_key generation failed %d time(s)', self._retries)
+            e = ConnectionError('auth_key generation failed {} time(s)'.format(self._retries))
             await self._disconnect(error=e)
             raise e
 
+        loop = asyncio.get_event_loop()
         self._log.debug('Starting send loop')
-        self._send_loop_handle = self._loop.create_task(self._send_loop())
+        self._send_loop_handle = loop.create_task(self._send_loop())
 
         self._log.debug('Starting receive loop')
-        self._recv_loop_handle = self._loop.create_task(self._recv_loop())
+        self._recv_loop_handle = loop.create_task(self._recv_loop())
 
         # _disconnected only completes after manual disconnection
         # or errors after which the sender cannot continue such
         # as failing to reconnect or any unexpected error.
         if self._disconnected.done():
-            self._disconnected = self._loop.create_future()
+            self._disconnected = loop.create_future()
 
         self._log.info('Connection to %s complete!', self._connection)
 
@@ -325,7 +338,7 @@ class MTProtoSender:
         """
         Cleanly disconnects and then reconnects.
         """
-        self._log.debug('Closing current connection...')
+        self._log.info('Closing current connection to begin reconnect...')
         await self._connection.disconnect()
 
         await helpers._cancel(
@@ -345,15 +358,31 @@ class MTProtoSender:
         self._state.reset()
 
         retries = self._retries if self._auto_reconnect else 0
-        for attempt in retry_range(retries):
+        
+        attempt = 0
+        ok = True
+        # We're already "retrying" to connect, so we don't want to force retries
+        for attempt in retry_range(retries, force_retry=False):
             try:
                 await self._connect()
             except (IOError, asyncio.TimeoutError) as e:
                 last_error = e
                 self._log.info('Failed reconnection attempt %d with %s',
                                attempt, e.__class__.__name__)
-
                 await asyncio.sleep(self._delay)
+            except BufferError as e:
+                # TODO there should probably only be one place to except all these errors
+                if isinstance(e, InvalidBufferError) and e.code == 404:
+                    self._log.info('Broken authorization key; resetting')
+                    self.auth_key.key = None
+                    if self._auth_key_callback:
+                        self._auth_key_callback(None)
+
+                    ok = False
+                    break
+                else:
+                    self._log.warning('Invalid buffer %s', e)
+
             except Exception as e:
                 last_error = e
                 self._log.exception('Unexpected exception reconnecting on '
@@ -365,10 +394,13 @@ class MTProtoSender:
                 self._pending_state.clear()
 
                 if self._auto_reconnect_callback:
-                    self._loop.create_task(self._auto_reconnect_callback())
+                    asyncio.get_event_loop().create_task(self._auto_reconnect_callback())
 
                 break
         else:
+            ok = False
+
+        if not ok:
             self._log.error('Automatic reconnection failed %d time(s)', attempt)
             await self._disconnect(error=last_error.with_traceback(None))
 
@@ -385,7 +417,7 @@ class MTProtoSender:
             # gets stuck.
             # TODO It still gets stuck? Investigate where and why.
             self._reconnecting = True
-            self._loop.create_task(self._reconnect(error))
+            asyncio.get_event_loop().create_task(self._reconnect(error))
 
     # Loops
 
@@ -398,7 +430,7 @@ class MTProtoSender:
         """
         while self._user_connected and not self._reconnecting:
             if self._pending_ack:
-                ack = RequestState(MsgsAck(list(self._pending_ack)), self._loop)
+                ack = RequestState(MsgsAck(list(self._pending_ack)))
                 self._send_queue.append(ack)
                 self._last_acks.append(ack)
                 self._pending_ack.clear()
@@ -416,13 +448,12 @@ class MTProtoSender:
                             len(batch), len(data))
 
             data = self._state.encrypt_message_data(data)
-            try:
-                await self._connection.send(data)
-            except IOError as e:
-                self._log.info('Connection closed while sending data')
-                self._start_reconnect(e)
-                return
 
+            # Whether sending succeeds or not, the popped requests are now
+            # pending because they're removed from the queue. If a reconnect
+            # occurs, they will be removed from pending state and re-enqueued
+            # so even if the network fails they won't be lost. If they were
+            # never re-enqueued, the future waiting for a response "locks".
             for state in batch:
                 if not isinstance(state, list):
                     if isinstance(state.request, TLRequest):
@@ -432,6 +463,13 @@ class MTProtoSender:
                         if isinstance(s.request, TLRequest):
                             self._pending_state[s.msg_id] = s
 
+            try:
+                await self._connection.send(data)
+            except IOError as e:
+                self._log.info('Connection closed while sending data')
+                self._start_reconnect(e)
+                return
+
             self._log.debug('Encrypted messages put in a queue to be sent')
 
     async def _recv_loop(self):
@@ -466,14 +504,14 @@ class MTProtoSender:
             except BufferError as e:
                 if isinstance(e, InvalidBufferError) and e.code == 404:
                     self._log.info('Broken authorization key; resetting')
+                    self.auth_key.key = None
+                    if self._auth_key_callback:
+                        self._auth_key_callback(None)
+
+                    await self._disconnect(error=e)
                 else:
                     self._log.warning('Invalid buffer %s', e)
-
-                self.auth_key.key = None
-                if self._auth_key_callback:
-                    self._auth_key_callback(None)
-
-                self._start_reconnect(e)
+                    self._start_reconnect(e)
                 return
             except Exception as e:
                 self._log.exception('Unhandled error while receiving data')
@@ -551,7 +589,7 @@ class MTProtoSender:
         if rpc_result.error:
             error = rpc_message_to_error(rpc_result.error, state.request)
             self._send_queue.append(
-                RequestState(MsgsAck([state.msg_id]), loop=self._loop))
+                RequestState(MsgsAck([state.msg_id])))
 
             if not state.future.cancelled():
                 state.future.set_exception(error)
@@ -738,8 +776,8 @@ class MTProtoSender:
         enqueuing a :tl:`MsgsStateInfo` to be sent at a later point.
         """
         self._send_queue.append(RequestState(MsgsStateInfo(
-            req_msg_id=message.msg_id, info=chr(1) * len(message.obj.msg_ids)),
-            loop=self._loop))
+            req_msg_id=message.msg_id, info=chr(1) * len(message.obj.msg_ids)
+        )))
 
     async def _handle_msg_all(self, message):
         """
