@@ -21,10 +21,6 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
     have access to all their sender and chat properties and methods.
 
     Members:
-        id (`int`):
-            The ID of this message. This field is *always* present.
-            Any other member is optional and may be `None`.
-
         out (`bool`):
             Whether the message is outgoing (i.e. you sent it from
             another session) or incoming (i.e. someone else sent it).
@@ -44,7 +40,6 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             Whether you have read the media in this message
             or not, e.g. listened to the voice note media.
 
-
         silent (`bool`):
             Whether the message should notify people with sound or not.
             Previously used in channels, but since 9 August 2019, it can
@@ -62,11 +57,38 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         legacy (`bool`):
             Whether this is a legacy message or not.
 
-        to_id (:tl:`Peer`):
+        edit_hide (`bool`):
+            Whether the edited mark of this message is edited
+            should be hidden (e.g. in GUI clients) or shown.
+
+        pinned (`bool`):
+            Whether this message is currently pinned or not.
+
+        id (`int`):
+            The ID of this message. This field is *always* present.
+            Any other member is optional and may be `None`.
+
+        from_id (:tl:`Peer`):
+            The peer who sent this message, which is either
+            :tl:`PeerUser`, :tl:`PeerChat` or :tl:`PeerChannel`.
+            This value will be `None` for anonymous messages.
+
+        peer_id (:tl:`Peer`):
             The peer to which this message was sent, which is either
             :tl:`PeerUser`, :tl:`PeerChat` or :tl:`PeerChannel`. This
             will always be present except for empty messages.
 
+        fwd_from (:tl:`MessageFwdHeader`):
+            The original forward header if this message is a forward.
+            You should probably use the `forward` property instead.
+
+        via_bot_id (`int`):
+            The ID of the bot used to send this message
+            through its inline mode (e.g. "via @like").
+
+        reply_to (:tl:`MessageReplyHeader`):
+            The original reply header if this message is replying to another.
+
         date (`datetime`):
             The UTC+0 `datetime` object indicating when this message
             was sent. This will always be present except for empty
@@ -77,25 +99,6 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             <telethon.tl.custom.message.Message>` instances,
             which will be `None` for other types of messages.
 
-        action (:tl:`MessageAction`):
-            The message action object of the message for :tl:`MessageService`
-            instances, which will be `None` for other types of messages.
-
-        from_id (`int`):
-            The ID of the user who sent this message. This will be
-            `None` if the message was sent in a broadcast channel.
-
-        reply_to_msg_id (`int`):
-            The ID to which this message is replying to, if any.
-
-        fwd_from (:tl:`MessageFwdHeader`):
-            The original forward header if this message is a forward.
-            You should probably use the `forward` property instead.
-
-        via_bot_id (`int`):
-            The ID of the bot used to send this message
-            through its inline mode (e.g. "via @like").
-
         media (:tl:`MessageMedia`):
             The media sent with this message if any (such as
             photos, videos, documents, gifs, stickers, etc.).
@@ -119,13 +122,15 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             The number of views this message from a broadcast
             channel has. This is also present in forwards.
 
+        forwards (`int`):
+            The number of times this message has been forwarded.
+
+        replies (`int`):
+            The number of times another message has replied to this message.
+
         edit_date (`datetime`):
             The date when this message was last edited.
 
-        edit_hide (`bool`):
-            Whether the edited mark of this message is edited
-            should be hidden (e.g. in GUI clients) or shown.
-
         post_author (`str`):
             The display name of the message sender to
             show in messages sent to broadcast channels.
@@ -138,6 +143,10 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         restriction_reason (List[:tl:`RestrictionReason`])
             An optional list of reasons why this message was restricted.
             If the list is `None`, this message has not been restricted.
+
+        action (:tl:`MessageAction`):
+            The message action object of the message for :tl:`MessageService`
+            instances, which will be `None` for other types of messages.
     """
 
     # region Initialization
@@ -147,11 +156,11 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             self, id,
 
             # Common to Message and MessageService (mandatory)
-            to_id=None, date=None,
+            peer_id=None, date=None,
 
             # Common to Message and MessageService (flags)
             out=None, mentioned=None, media_unread=None, silent=None,
-            post=None, from_id=None, reply_to_msg_id=None,
+            post=None, from_id=None, reply_to=None,
 
             # For Message (mandatory)
             message=None,
@@ -160,36 +169,38 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
             fwd_from=None, via_bot_id=None, media=None, reply_markup=None,
             entities=None, views=None, edit_date=None, post_author=None,
             grouped_id=None, from_scheduled=None, legacy=None,
-            edit_hide=None, restriction_reason=None,
+            edit_hide=None, pinned=None, restriction_reason=None,
+            forwards=None, replies=None,
 
             # For MessageAction (mandatory)
             action=None):
-        # Common properties to all messages
-        self.id = id
-        self.to_id = to_id
-        self.date = date
+        # Common properties to messages, then to service (in the order they're defined in the `.tl`)
         self.out = out
         self.mentioned = mentioned
         self.media_unread = media_unread
         self.silent = silent
         self.post = post
-        self.from_id = from_id
-        self.reply_to_msg_id = reply_to_msg_id
-        self.message = message
-        self.fwd_from = fwd_from
-        self.via_bot_id = via_bot_id
-        self.media = None if isinstance(
-            media, types.MessageMediaEmpty) else media
-
-        self.reply_markup = reply_markup
-        self.entities = entities
-        self.views = views
-        self.edit_date = edit_date
-        self.post_author = post_author
-        self.grouped_id = grouped_id
         self.from_scheduled = from_scheduled
         self.legacy = legacy
         self.edit_hide = edit_hide
+        self.id = id
+        self.from_id = from_id
+        self.peer_id = peer_id
+        self.fwd_from = fwd_from
+        self.via_bot_id = via_bot_id
+        self.reply_to = reply_to
+        self.date = date
+        self.message = message
+        self.media =  None if isinstance(media, types.MessageMediaEmpty) else media
+        self.reply_markup = reply_markup
+        self.entities = entities
+        self.views = views
+        self.forwards = forwards
+        self.replies = replies
+        self.edit_date = edit_date
+        self.pinned = pinned
+        self.post_author = post_author
+        self.grouped_id = grouped_id
         self.restriction_reason = restriction_reason
         self.action = action
 
@@ -206,20 +217,20 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         self._via_input_bot = None
         self._action_entities = None
 
-        if not out and isinstance(to_id, types.PeerUser):
-            chat_peer = types.PeerUser(from_id)
-            if from_id == to_id.user_id:
-                self.out = not self.fwd_from  # Patch out in our chat
-        else:
-            chat_peer = to_id
+        sender_id = None
+        if from_id is not None:
+            sender_id = utils.get_peer_id(from_id)
+        elif peer_id:
+            # If the message comes from a Channel, let the sender be it
+            # ...or...
+            # incoming messages in private conversations no longer have from_id
+            # (layer 119+), but the sender can only be the chat we're in.
+            if post or (not out and isinstance(peer_id, types.PeerUser)):
+                sender_id = utils.get_peer_id(peer_id)
 
         # Note that these calls would reset the client
-        ChatGetter.__init__(self, chat_peer, broadcast=post)
-        SenderGetter.__init__(self, from_id)
-
-        if post and not from_id and chat_peer:
-            # If the message comes from a Channel, let the sender be it
-            self._sender_id = utils.get_peer_id(chat_peer)
+        ChatGetter.__init__(self, peer_id, broadcast=post)
+        SenderGetter.__init__(self, sender_id)
 
         self._forward = None
 
@@ -326,10 +337,10 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         `True` if the message is a reply to some other message.
 
         Remember that you can access the ID of the message
-        this one is replying to through `reply_to_msg_id`,
+        this one is replying to through `reply_to.reply_to_msg_id`,
         and the `Message` object with `get_reply_message()`.
         """
-        return bool(self.reply_to_msg_id)
+        return self.reply_to is not None
 
     @property
     def forward(self):
@@ -552,6 +563,14 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         if isinstance(self.media, types.MessageMediaVenue):
             return self.media
 
+    @property
+    def dice(self):
+        """
+        The :tl:`MessageMediaDice` in this message, if it's a dice roll.
+        """
+        if isinstance(self.media, types.MessageMediaDice):
+            return self.media
+
     @property
     def action_entities(self):
         """
@@ -585,6 +604,27 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         """
         return self._via_input_bot
 
+    @property
+    def reply_to_msg_id(self):
+        """
+        Returns the message ID this message is replying to, if any.
+        This is equivalent to accessing ``.reply_to.reply_to_msg_id``.
+        """
+        return self.reply_to.reply_to_msg_id if self.reply_to else None
+
+    @property
+    def to_id(self):
+        """
+        Returns the peer to which this message was sent to. This used to exist
+        to infer the ``.peer_id``.
+        """
+        # If the client wasn't set we can't emulate the behaviour correctly,
+        # so as a best-effort simply return the chat peer.
+        if self._client and not self.out and self.is_private:
+            return types.PeerUser(self._client._self_id)
+
+        return self.peer_id
+
     # endregion Public Properties
 
     # region Public Methods
@@ -636,7 +676,7 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         The result will be cached after its first use.
         """
         if self._reply_message is None and self._client:
-            if not self.reply_to_msg_id:
+            if not self.reply_to:
                 return None
 
             # Bots cannot access other bots' messages by their ID.
@@ -652,7 +692,7 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
                 # directly by its ID.
                 self._reply_message = await self._client.get_messages(
                     self._input_chat if self.is_channel else None,
-                    ids=self.reply_to_msg_id
+                    ids=self.reply_to.reply_to_msg_id
                 )
 
         return self._reply_message
@@ -753,19 +793,24 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         with the ``message`` already set.
         """
         if self._client:
+            # Passing the entire message is important, in case it has to be
+            # refetched for a fresh file reference.
             return await self._client.download_media(self, *args, **kwargs)
 
     async def click(self, i=None, j=None,
-                    *, text=None, filter=None, data=None):
+                    *, text=None, filter=None, data=None, share_phone=None,
+                    share_geo=None):
         """
-        Calls `button.click <telethon.tl.custom.messagebutton.MessageButton.click>`
+        Calls :tl:`SendVote` with the specified poll option
+        or `button.click <telethon.tl.custom.messagebutton.MessageButton.click>`
         on the specified button.
 
-        Does nothing if the message has no buttons.
+        Does nothing if the message is not a poll or has no buttons.
 
         Args:
-            i (`int`):
-                Clicks the i'th button (starting from the index 0).
+            i (`int` | `list`):
+                Clicks the i'th button or poll option (starting from the index 0).
+                For multiple-choice polls, a list with the indices should be used.
                 Will ``raise IndexError`` if out of bounds. Example:
 
                 >>> message = ...  # get the message somehow
@@ -788,15 +833,21 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
                 This is equivalent to ``message.buttons[0][1].click()``.
 
             text (`str` | `callable`):
-                Clicks the first button with the text "text". This may
+                Clicks the first button or poll option with the text "text". This may
                 also be a callable, like a ``re.compile(...).match``,
                 and the text will be passed to it.
 
+                If you need to select multiple options in a poll,
+                pass a list of indices to the ``i`` parameter.
+
             filter (`callable`):
-                Clicks the first button for which the callable
+                Clicks the first button or poll option for which the callable
                 returns `True`. The callable should accept a single
                 `MessageButton <telethon.tl.custom.messagebutton.MessageButton>`
-                argument.
+                or `PollAnswer <telethon.tl.types.PollAnswer>` argument.
+
+                If you need to select multiple options in a poll,
+                pass a list of indices to the ``i`` parameter.
 
             data (`bytes`):
                 This argument overrides the rest and will not search any
@@ -805,6 +856,28 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
                 that if the message does not have this data, it will
                 ``raise DataInvalidError``.
 
+            share_phone (`bool` | `str` | tl:`InputMediaContact`):
+                When clicking on a keyboard button requesting a phone number
+                (:tl:`KeyboardButtonRequestPhone`), this argument must be
+                explicitly set to avoid accidentally sharing the number.
+
+                It can be `True` to automatically share the current user's
+                phone, a string to share a specific phone number, or a contact
+                media to specify all details.
+
+                If the button is pressed without this, `ValueError` is raised.
+
+            share_geo (`tuple` | `list` | tl:`InputMediaGeoPoint`):
+                When clicking on a keyboard button requesting a geo location
+                (:tl:`KeyboardButtonRequestGeoLocation`), this argument must
+                be explicitly set to avoid accidentally sharing the location.
+
+                It must be a `tuple` of `float` as ``(longitude, latitude)``,
+                or a :tl:`InputGeoPoint` instance to avoid accidentally using
+                the wrong roder.
+
+                If the button is pressed without this, `ValueError` is raised.
+
             Example:
 
                 .. code-block:: python
@@ -820,6 +893,9 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
 
                     # Click by data
                     await message.click(data=b'payload')
+
+                    # Click on a button requesting a phone
+                    await message.click(0, share_phone=True)
         """
         if not self._client:
             return
@@ -836,38 +912,80 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
                         data=data
                     )
                 )
-            except errors.BotTimeout:
+            except errors.BotResponseTimeoutError:
                 return None
 
         if sum(int(x is not None) for x in (i, text, filter)) >= 2:
             raise ValueError('You can only set either of i, text or filter')
 
+        # Finding the desired poll options and sending them
+        if self.poll is not None:
+            def find_options():
+                answers = self.poll.poll.answers
+                if i is not None:
+                    if utils.is_list_like(i):
+                        return [answers[idx].option for idx in i]
+                    return [answers[i].option]
+                if text is not None:
+                    if callable(text):
+                        for answer in answers:
+                            if text(answer.text):
+                                return [answer.option]
+                    else:
+                        for answer in answers:
+                            if answer.text == text:
+                                return [answer.option]
+                    return
+
+                if filter is not None:
+                    for answer in answers:
+                        if filter(answer):
+                            return [answer.option]
+                    return
+
+            options = find_options()
+            if options is None:
+                return
+            return await self._client(
+                functions.messages.SendVoteRequest(
+                    peer=self._input_chat,
+                    msg_id=self.id,
+                    options=options
+                )
+            )
+
         if not await self.get_buttons():
             return  # Accessing the property sets self._buttons[_flat]
 
-        if text is not None:
-            if callable(text):
+        def find_button():
+            nonlocal i
+            if text is not None:
+                if callable(text):
+                    for button in self._buttons_flat:
+                        if text(button.text):
+                            return button
+                else:
+                    for button in self._buttons_flat:
+                        if button.text == text:
+                            return button
+                return
+
+            if filter is not None:
                 for button in self._buttons_flat:
-                    if text(button.text):
-                        return await button.click()
+                    if filter(button):
+                        return button
+                return
+
+            if i is None:
+                i = 0
+            if j is None:
+                return self._buttons_flat[i]
             else:
-                for button in self._buttons_flat:
-                    if button.text == text:
-                        return await button.click()
-            return
+                return self._buttons[i][j]
 
-        if filter is not None:
-            for button in self._buttons_flat:
-                if filter(button):
-                    return await button.click()
-            return
-
-        if i is None:
-            i = 0
-        if j is None:
-            return await self._buttons_flat[i].click()
-        else:
-            return await self._buttons[i][j].click()
+        button = find_button()
+        if button:
+            return await button.click(share_phone=share_phone, share_geo=share_geo)
 
     async def mark_read(self):
         """
@@ -890,9 +1008,19 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         #      maybe just make it illegal to call messages from raw API?
         #      That or figure out a way to always set it directly.
         if self._client:
-            await self._client.pin_message(
+            return await self._client.pin_message(
                 await self.get_input_chat(), self.id, notify=notify)
 
+    async def unpin(self):
+        """
+        Unpins the message. Shorthand for
+        `telethon.client.messages.MessageMethods.unpin_message`
+        with both ``entity`` and ``message`` already set.
+        """
+        if self._client:
+            return await self._client.unpin_message(
+                await self.get_input_chat(), self.id)
+
     # endregion Public Methods
 
     # region Private Methods
@@ -952,7 +1080,8 @@ class Message(ChatGetter, SenderGetter, TLObject, abc.ABC):
         for row in self.reply_markup.rows:
             for button in row.buttons:
                 if isinstance(button, types.KeyboardButtonSwitchInline):
-                    if button.same_peer:
+                    # no via_bot_id means the bot sent the message itself (#1619)
+                    if button.same_peer or not self.via_bot_id:
                         bot = self.input_sender
                         if not bot:
                             raise ValueError('No input sender')
